const std = @import("std");
const leapc = @import("leapc/leapc.zig");
const signal = @cImport({
    @cInclude("signal.h");
});

const hand_fmt = "Hand with id {d}, handedness: {s}\n";

var keep_running: bool = true;

pub fn main() !void {
    _ = signal.signal(signal.SIGINT, &sigintHandler);

    var handle: leapc.LEAP_CONNECTION = null;

    var config: leapc.LEAP_CONNECTION_CONFIG = std.mem.zeroes(leapc.LEAP_CONNECTION_CONFIG);
    config.size = @sizeOf(leapc.LEAP_CONNECTION_CONFIG);

    // Create connection
    var result: leapc.eLeapRS = leapc.LeapCreateConnection(null, &handle);
    if (result != leapc.eLeapRS_Success) {
        std.debug.print("LeapCreateConnection() error {d}\n", .{result});
        return error.LeapCreateConnectionError;
    }
    defer leapc.LeapDestroyConnection(handle);

    // Open connection
    result = leapc.LeapOpenConnection(handle);
    if (result != leapc.eLeapRS_Success) {
        std.debug.print("LeapOpenConnection() error {d}\n", .{result});
        return error.LeapOpenConnectionError;
    }
    defer leapc.LeapCloseConnection(handle);

    // Set allocator
    const allocator: leapc.LEAP_ALLOCATOR = .{
        .allocate = allocate,
        .deallocate = deallocate,
        .state = null,
    };
    result = leapc.LeapSetAllocator(handle, &allocator);
    if (result != leapc.eLeapRS_Success) {
        std.debug.print("LeapSetAllocator() error {d}\n", .{result});
        return error.LeapSetAllocatorError;
    }

    var message: leapc.LEAP_CONNECTION_MESSAGE = std.mem.zeroInit(leapc.LEAP_CONNECTION_MESSAGE, .{});

    while (keep_running) {
        result = leapc.LeapPollConnection(handle, 100, &message);

        if (result != leapc.eLeapRS_Success and result != leapc.eLeapRS_Timeout) {
            std.debug.print("LeapPollConnection() error {d}\n", .{result});
            return error.LeapPollConnectionError;
        }

        if (result == leapc.eLeapRS_Timeout) {
            continue;
        }

        switch (message.type) {
            leapc.eLeapEventType_Connection => {
                std.debug.print("Connected!\n", .{});
            },
            leapc.eLeapEventType_Tracking => {

                // Those are equivalent, you can grab the tracking event directly from the unnamed union,
                // or cast the pointer found in the union to whatever event you have
                const frame: *const leapc.LEAP_TRACKING_EVENT = message.unnamed_0.tracking_event;
                //const frame: *const leapc.LEAP_TRACKING_EVENT = @as(*const leapc.LEAP_TRACKING_EVENT, @ptrCast((@alignCast(message.unnamed_0.pointer))));

                std.debug.print("Frame id {d}, {d} hands found. FPS: {d}\n", .{ frame.tracking_frame_id, frame.nHands, frame.framerate });

                for (0..frame.nHands) |i| {
                    const hand: leapc.LEAP_HAND = frame.pHands[i];
                    std.debug.print(hand_fmt, .{ hand.id, if (hand.type == leapc.eLeapHandType_Left) "Left" else "Right" });
                    std.debug.print("Palm position: ({d}, {d}, {d})\n", .{
                        hand.palm.position.unnamed_0.unnamed_0.x,
                        hand.palm.position.unnamed_0.unnamed_0.y,
                        hand.palm.position.unnamed_0.unnamed_0.z,
                    });
                }
            },
            else => {
                std.debug.print("Message unsupported for now\n", .{});
            },
        }
    }
}

/// Handler for SIGINT (sets `keep_running` to true to stop the main loop)
fn sigintHandler(_: c_int) callconv(.C) void {
    std.debug.print("Received a sigint\n", .{});
    keep_running = false;
}

/// Allocate memory for LeapC using a page allocator
fn allocate(size: u32, typeHint: leapc.eLeapAllocatorType, _: ?*anyopaque) callconv(.C) ?*anyopaque {
    const allocator = std.heap.page_allocator;
    switch (typeHint) {
        leapc.eLeapAllocatorType_Int8 => {
            const memory = allocator.alloc(i8, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_Int16 => {
            const memory = allocator.alloc(i16, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_Int32 => {
            const memory = allocator.alloc(i32, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_Int64 => {
            const memory = allocator.alloc(i64, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_Uint8 => {
            const memory = allocator.alloc(u8, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_UInt16 => {
            const memory = allocator.alloc(u16, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_UInt32 => {
            const memory = allocator.alloc(u32, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_UInt64 => {
            const memory = allocator.alloc(u64, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_Float => {
            const memory = allocator.alloc(f32, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        leapc.eLeapAllocatorType_Double => {
            const memory = allocator.alloc(f64, size) catch {
                return null;
            };
            return @ptrCast(memory);
        },
        else => {
            return null;
        },
    }
}

/// Deallocate memory for LeapC
fn deallocate(_: ?*anyopaque, _: ?*anyopaque) callconv(.C) void {
    // How to deallocate properly if we don't know what type we need to cast it in?
    // We could add the type to the state that is being sent to the allocator and deallocator?
}
